# mp_limb_t mulredc1_7(mp_limb_t * z, const mp_limb_t x, const mp_limb_t * y,
#                 const mp_limb_t *m, mp_limb_t inv_m);
#
# Linux:   z: %rdi, x: %rsi, y: %rdx, m: %rcx, inv_m: %r8
#          Needs %rbx, %rsp, %rbp, %r12-%r15 restored
# Windows: z: %rcx, x: %rdx, y: %r8,  m: %r9, inv_m: 28(%rsp)
#          Needs %rbx, %rbp, %rdi, %rsi, %r12...%15 restored



















	.text
.p2align 6 # Opteron L1 code cache line is 64 bytes long
	.globl mulredc1_7
	.type	mulredc1_7,@function

# Implements multiplication and REDC for one input numbers of LENGTH words
# and a multiplier of one word
# Uses Linux ABI

# Values that are referenced only once in the loop over j go into r8 .. r14,
# In the inner loop (over j), tmp, x[i], y, m, and u are constant.
# tmp[j], tmp[j+1], tmp[j+2] are updated frequently. These 8 values
# stay in registers and are referenced as
# YP = y, MP = m, 
# X = x, T0 = tmp[j], T1 = tmp[j+1], CY = carry


# Register vars: T0 = %rsi, T1 = %rbx, CY = %rcx, X = %r14, U = %r11
#                YP = %r9, MP = %r10

mulredc1_7:


#########################################################################
# i = 0 pass
#########################################################################

# register values at loop entry: %r9 = y, %r10 = m

# We need to compute u

	movq	(%rdx), %rax		# rax = y[0] (time critical, do first)
	pushq	%rbx
	pushq	%r14
	movq	%rdx, %r9
	movq	%rcx, %r10
	movq    %rsi, %r14		# store x in X
	# ZP is same as passed in


	xorl	%ecx, %ecx		# set %CY to 0

	mulq	%r14			# rdx:rax = y[0] * x

	movq 	%rax, %rsi		# Move low word of product to T0
	movq	%rdx, %rbx		# Move high word of product to T1

	imulq	%r8, %rax	# %rax = ((x[i]*y[0]+tmp[0])*invm)%2^64
	movq	%rax, %r11			# this is the new u value

	mulq	(%r10)			# multipy u*m[0]
	addq	%rax, %rsi		# Now %T0 = 0, need not be stored
	movq	8(%r9), %rax		# Fetch y[1]
	adcq	%rdx, %rbx		# 
	setc	%cl
	# CY:T1:T0 <= 2*(2^64-1)^2 <= 2^2*128 - 4*2^64 + 2, hence
	# CY:T1 <= 2*2^64 - 4

# Now T0 = %rbx, T1 = %rsi


# Pass for j = 1
# Register values at entry: 
# %rax = y[j], X = x, U = u
# T0 = value to store in tmp[j], T1 undefined 
# CY = carry into T1 (is <= 2)
# We have CY:T1 <= 2 * 2^64 - 2

	movq	%rcx, %rsi		# T1 = CY <= 1

	# Here, T1:T0 <= 2*2^64 - 2
	mulq	%r14		# y[j] * x
	# rdx:rax <= (2^64-1)^2 <= 2^128 - 2*2^64 + 1
	addq	%rax, %rbx	# Add low word to T0
	movq	8(%r10), %rax	# Fetch m[j] into %rax
	adcq	%rdx, %rsi	# Add high word with carry to T1
	# T1:T0 <= 2^128 - 2*2^64 + 1 + 2*2^64 - 2 <= 2^128 - 1, no carry!
	
	mulq	%r11		# m[j]*u
	# rdx:rax <= 2^128 - 2*2^64 + 1, T1:T0 <= 2^128 - 1
	addq	%rbx, %rax	# Add T0 and low word
	movq	%rax, 0(%rdi)	# Store T0 in z[1-1]
	movq	16(%r9), %rax	# Fetch y[j+1] = y[2] into %rax
	adcq	%rdx, %rsi	# Add high word with carry to T1
	setc	%cl		# CY <= 1
	# CY:T1:T0 <= 2^128 - 1 + 2^128 - 2*2^64 + 1 <=
	#             2 * 2^128 - 2*2^64 ==> CY:T1 <= 2 * 2^64 - 2

# Now T0 = %rsi, T1 = %rbx


# Pass for j = 2
# Register values at entry: 
# %rax = y[j], X = x, U = u
# T0 = value to store in tmp[j], T1 undefined 
# CY = carry into T1 (is <= 2)
# We have CY:T1 <= 2 * 2^64 - 2

	movq	%rcx, %rbx		# T1 = CY <= 1

	# Here, T1:T0 <= 2*2^64 - 2
	mulq	%r14		# y[j] * x
	# rdx:rax <= (2^64-1)^2 <= 2^128 - 2*2^64 + 1
	addq	%rax, %rsi	# Add low word to T0
	movq	16(%r10), %rax	# Fetch m[j] into %rax
	adcq	%rdx, %rbx	# Add high word with carry to T1
	# T1:T0 <= 2^128 - 2*2^64 + 1 + 2*2^64 - 2 <= 2^128 - 1, no carry!
	
	mulq	%r11		# m[j]*u
	# rdx:rax <= 2^128 - 2*2^64 + 1, T1:T0 <= 2^128 - 1
	addq	%rsi, %rax	# Add T0 and low word
	movq	%rax, 8(%rdi)	# Store T0 in z[2-1]
	movq	24(%r9), %rax	# Fetch y[j+1] = y[3] into %rax
	adcq	%rdx, %rbx	# Add high word with carry to T1
	setc	%cl		# CY <= 1
	# CY:T1:T0 <= 2^128 - 1 + 2^128 - 2*2^64 + 1 <=
	#             2 * 2^128 - 2*2^64 ==> CY:T1 <= 2 * 2^64 - 2

# Now T0 = %rbx, T1 = %rsi


# Pass for j = 3
# Register values at entry: 
# %rax = y[j], X = x, U = u
# T0 = value to store in tmp[j], T1 undefined 
# CY = carry into T1 (is <= 2)
# We have CY:T1 <= 2 * 2^64 - 2

	movq	%rcx, %rsi		# T1 = CY <= 1

	# Here, T1:T0 <= 2*2^64 - 2
	mulq	%r14		# y[j] * x
	# rdx:rax <= (2^64-1)^2 <= 2^128 - 2*2^64 + 1
	addq	%rax, %rbx	# Add low word to T0
	movq	24(%r10), %rax	# Fetch m[j] into %rax
	adcq	%rdx, %rsi	# Add high word with carry to T1
	# T1:T0 <= 2^128 - 2*2^64 + 1 + 2*2^64 - 2 <= 2^128 - 1, no carry!
	
	mulq	%r11		# m[j]*u
	# rdx:rax <= 2^128 - 2*2^64 + 1, T1:T0 <= 2^128 - 1
	addq	%rbx, %rax	# Add T0 and low word
	movq	%rax, 16(%rdi)	# Store T0 in z[3-1]
	movq	32(%r9), %rax	# Fetch y[j+1] = y[4] into %rax
	adcq	%rdx, %rsi	# Add high word with carry to T1
	setc	%cl		# CY <= 1
	# CY:T1:T0 <= 2^128 - 1 + 2^128 - 2*2^64 + 1 <=
	#             2 * 2^128 - 2*2^64 ==> CY:T1 <= 2 * 2^64 - 2

# Now T0 = %rsi, T1 = %rbx


# Pass for j = 4
# Register values at entry: 
# %rax = y[j], X = x, U = u
# T0 = value to store in tmp[j], T1 undefined 
# CY = carry into T1 (is <= 2)
# We have CY:T1 <= 2 * 2^64 - 2

	movq	%rcx, %rbx		# T1 = CY <= 1

	# Here, T1:T0 <= 2*2^64 - 2
	mulq	%r14		# y[j] * x
	# rdx:rax <= (2^64-1)^2 <= 2^128 - 2*2^64 + 1
	addq	%rax, %rsi	# Add low word to T0
	movq	32(%r10), %rax	# Fetch m[j] into %rax
	adcq	%rdx, %rbx	# Add high word with carry to T1
	# T1:T0 <= 2^128 - 2*2^64 + 1 + 2*2^64 - 2 <= 2^128 - 1, no carry!
	
	mulq	%r11		# m[j]*u
	# rdx:rax <= 2^128 - 2*2^64 + 1, T1:T0 <= 2^128 - 1
	addq	%rsi, %rax	# Add T0 and low word
	movq	%rax, 24(%rdi)	# Store T0 in z[4-1]
	movq	40(%r9), %rax	# Fetch y[j+1] = y[5] into %rax
	adcq	%rdx, %rbx	# Add high word with carry to T1
	setc	%cl		# CY <= 1
	# CY:T1:T0 <= 2^128 - 1 + 2^128 - 2*2^64 + 1 <=
	#             2 * 2^128 - 2*2^64 ==> CY:T1 <= 2 * 2^64 - 2

# Now T0 = %rbx, T1 = %rsi


# Pass for j = 5
# Register values at entry: 
# %rax = y[j], X = x, U = u
# T0 = value to store in tmp[j], T1 undefined 
# CY = carry into T1 (is <= 2)
# We have CY:T1 <= 2 * 2^64 - 2

	movq	%rcx, %rsi		# T1 = CY <= 1

	# Here, T1:T0 <= 2*2^64 - 2
	mulq	%r14		# y[j] * x
	# rdx:rax <= (2^64-1)^2 <= 2^128 - 2*2^64 + 1
	addq	%rax, %rbx	# Add low word to T0
	movq	40(%r10), %rax	# Fetch m[j] into %rax
	adcq	%rdx, %rsi	# Add high word with carry to T1
	# T1:T0 <= 2^128 - 2*2^64 + 1 + 2*2^64 - 2 <= 2^128 - 1, no carry!
	
	mulq	%r11		# m[j]*u
	# rdx:rax <= 2^128 - 2*2^64 + 1, T1:T0 <= 2^128 - 1
	addq	%rbx, %rax	# Add T0 and low word
	movq	%rax, 32(%rdi)	# Store T0 in z[5-1]
	movq	48(%r9), %rax	# Fetch y[j+1] = y[6] into %rax
	adcq	%rdx, %rsi	# Add high word with carry to T1
	setc	%cl		# CY <= 1
	# CY:T1:T0 <= 2^128 - 1 + 2^128 - 2*2^64 + 1 <=
	#             2 * 2^128 - 2*2^64 ==> CY:T1 <= 2 * 2^64 - 2

# Now T0 = %rsi, T1 = %rbx


# Pass for j = 6. Don't fetch new data from y[j+1].

	movq	%rcx, %rbx		# T1 = CY <= 1
	
	mulq	%r14		# y[j] * x[i]
	addq	%rax, %rsi	# Add low word to T0
	movq	48(%r10), %rax	# Fetch m[j] into %rax
	adcq	%rdx, %rbx 	# Add high word with carry to T1
	mulq    %r11		# m[j]*u
	addq	%rax, %rsi	# Add low word to T0
	movq	%rsi, 40(%rdi)	# Store T0 in z[j-1]
	adcq	%rdx, %rbx	# Add high word with carry to T1
	movq	%rbx, 48(%rdi)	# Store T1 in tmp[j]
	setc	%cl		# %CY <= 1

	movq	%rcx, %rax	# use carry as return value
 	popq	%r14
	popq	%rbx
	ret
